@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2006-2008 Edgewall Software
+# Copyright (C) 2006-2009 Edgewall Software
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -16,10 +16,6 @@ streams.
 """
 
 from itertools import chain
-try:
-    frozenset
-except NameError:
-    from sets import ImmutableSet as frozenset
 import re
 
 from genshi.core import escape, Attrs, Markup, Namespace, QName, StreamEventKind
@@ -30,6 +26,7 @@ __all__ = ['encode', 'get_serializer', 'DocType', 'XMLSerializer',
            'XHTMLSerializer', 'HTMLSerializer', 'TextSerializer']
 __docformat__ = 'restructuredtext en'
 
+
 def encode(iterator, method='xml', encoding='utf-8', out=None):
     """Encode serializer output into a string.
     
@@ -57,10 +54,11 @@ def encode(iterator, method='xml', encoding='utf-8', out=None):
     else:
         _encode = lambda string: string
     if out is None:
-        return _encode(u''.join(list(iterator)))
+        return _encode(''.join(list(iterator)))
     for chunk in iterator:
         out.write(_encode(chunk))
 
+
 def get_serializer(method='xml', **kwargs):
     """Return a serializer object for the given method.
     
@@ -133,6 +131,7 @@ class DocType(object):
     )
     SVG = SVG_FULL
 
+    @classmethod
     def get(cls, name):
         """Return the ``(name, pubid, sysid)`` tuple of the ``DOCTYPE``
         declaration for the specified name.
@@ -168,7 +167,6 @@ class DocType(object):
             'svg-basic': cls.SVG_BASIC,
             'svg-tiny': cls.SVG_TINY
         }.get(name.lower())
-    get = classmethod(get)
 
 
 class XMLSerializer(object):
@@ -176,14 +174,14 @@ class XMLSerializer(object):
     
     >>> from genshi.builder import tag
     >>> elem = tag.div(tag.a(href='foo'), tag.br, tag.hr(noshade=True))
-    >>> print ''.join(XMLSerializer()(elem.generate()))
+    >>> print(''.join(XMLSerializer()(elem.generate())))
     <div><a href="foo"/><br/><hr noshade="True"/></div>
     """
 
     _PRESERVE_SPACE = frozenset()
 
     def __init__(self, doctype=None, strip_whitespace=True,
-                 namespace_prefixes=None):
+                 namespace_prefixes=None, cache=True):
         """Initialize the XML serializer.
         
         :param doctype: a ``(name, pubid, sysid)`` tuple that represents the
@@ -192,42 +190,60 @@ class XMLSerializer(object):
                         defined in `DocType.get`
         :param strip_whitespace: whether extraneous whitespace should be
                                  stripped from the output
+        :param cache: whether to cache the text output per event, which
+                      improves performance for repetitive markup
         :note: Changed in 0.4.2: The  `doctype` parameter can now be a string.
+        :note: Changed in 0.6: The `cache` parameter was added
         """
         self.filters = [EmptyTagFilter()]
         if strip_whitespace:
             self.filters.append(WhitespaceFilter(self._PRESERVE_SPACE))
-        self.filters.append(NamespaceFlattener(prefixes=namespace_prefixes))
+        self.filters.append(NamespaceFlattener(prefixes=namespace_prefixes,
+                                               cache=cache))
         if doctype:
             self.filters.append(DocTypeInserter(doctype))
+        self.cache = cache
 
     def __call__(self, stream):
         have_decl = have_doctype = False
         in_cdata = False
 
+        cache = {}
+        cache_get = cache.get
+        if self.cache:
+            def _emit(kind, input, output):
+                cache[kind, input] = output
+                return output
+        else:
+            def _emit(kind, input, output):
+                return output
+
         for filter_ in self.filters:
             stream = filter_(stream)
         for kind, data, pos in stream:
+            cached = cache_get((kind, data))
+            if cached is not None:
+                yield cached
 
-            if kind is START or kind is EMPTY:
+            elif kind is START or kind is EMPTY:
                 tag, attrib = data
                 buf = ['<', tag]
                 for attr, value in attrib:
                     buf += [' ', attr, '="', escape(value), '"']
                 buf.append(kind is EMPTY and '/>' or '>')
-                yield Markup(u''.join(buf))
+                yield _emit(kind, data, Markup(''.join(buf)))
 
             elif kind is END:
-                yield Markup('</%s>' % data)
+                yield _emit(kind, data, Markup('</%s>' % data))
 
             elif kind is TEXT:
                 if in_cdata:
-                    yield data
+                    yield _emit(kind, data, data)
                 else:
-                    yield escape(data, quotes=False)
+                    yield _emit(kind, data, escape(data, quotes=False))
 
             elif kind is COMMENT:
-                yield Markup('<!--%s-->' % data)
+                yield _emit(kind, data, Markup('<!--%s-->' % data))
 
             elif kind is XML_DECL and not have_decl:
                 version, encoding, standalone = data
@@ -238,7 +254,7 @@ class XMLSerializer(object):
                     standalone = standalone and 'yes' or 'no'
                     buf.append(' standalone="%s"' % standalone)
                 buf.append('?>\n')
-                yield Markup(u''.join(buf))
+                yield Markup(''.join(buf))
                 have_decl = True
 
             elif kind is DOCTYPE and not have_doctype:
@@ -251,7 +267,7 @@ class XMLSerializer(object):
                 if sysid:
                     buf.append(' "%s"')
                 buf.append('>\n')
-                yield Markup(u''.join(buf)) % filter(None, data)
+                yield Markup(''.join(buf)) % tuple([p for p in data if p])
                 have_doctype = True
 
             elif kind is START_CDATA:
@@ -263,7 +279,7 @@ class XMLSerializer(object):
                 in_cdata = False
 
             elif kind is PI:
-                yield Markup('<?%s %s?>' % data)
+                yield _emit(kind, data, Markup('<?%s %s?>' % data))
 
 
 class XHTMLSerializer(XMLSerializer):
@@ -271,7 +287,7 @@ class XHTMLSerializer(XMLSerializer):
     
     >>> from genshi.builder import tag
     >>> elem = tag.div(tag.a(href='foo'), tag.br, tag.hr(noshade=True))
-    >>> print ''.join(XHTMLSerializer()(elem.generate()))
+    >>> print(''.join(XHTMLSerializer()(elem.generate())))
     <div><a href="foo"></a><br /><hr noshade="noshade" /></div>
     """
 
@@ -287,17 +303,19 @@ class XHTMLSerializer(XMLSerializer):
     ])
 
     def __init__(self, doctype=None, strip_whitespace=True,
-                 namespace_prefixes=None, drop_xml_decl=True):
+                 namespace_prefixes=None, drop_xml_decl=True, cache=True):
         super(XHTMLSerializer, self).__init__(doctype, False)
         self.filters = [EmptyTagFilter()]
         if strip_whitespace:
             self.filters.append(WhitespaceFilter(self._PRESERVE_SPACE))
         namespace_prefixes = namespace_prefixes or {}
         namespace_prefixes['http://www.w3.org/1999/xhtml'] = ''
-        self.filters.append(NamespaceFlattener(prefixes=namespace_prefixes))
+        self.filters.append(NamespaceFlattener(prefixes=namespace_prefixes,
+                                               cache=cache))
         if doctype:
             self.filters.append(DocTypeInserter(doctype))
         self.drop_xml_decl = drop_xml_decl
+        self.cache = cache
 
     def __call__(self, stream):
         boolean_attrs = self._BOOLEAN_ATTRS
@@ -306,19 +324,32 @@ class XHTMLSerializer(XMLSerializer):
         have_decl = have_doctype = False
         in_cdata = False
 
+        cache = {}
+        cache_get = cache.get
+        if self.cache:
+            def _emit(kind, input, output):
+                cache[kind, input] = output
+                return output
+        else:
+            def _emit(kind, input, output):
+                return output
+
         for filter_ in self.filters:
             stream = filter_(stream)
         for kind, data, pos in stream:
+            cached = cache_get((kind, data))
+            if cached is not None:
+                yield cached
 
-            if kind is START or kind is EMPTY:
+            elif kind is START or kind is EMPTY:
                 tag, attrib = data
                 buf = ['<', tag]
                 for attr, value in attrib:
                     if attr in boolean_attrs:
                         value = attr
-                    elif attr == u'xml:lang' and u'lang' not in attrib:
+                    elif attr == 'xml:lang' and 'lang' not in attrib:
                         buf += [' lang="', escape(value), '"']
-                    elif attr == u'xml:space':
+                    elif attr == 'xml:space':
                         continue
                     buf += [' ', attr, '="', escape(value), '"']
                 if kind is EMPTY:
@@ -328,19 +359,19 @@ class XHTMLSerializer(XMLSerializer):
                         buf.append('></%s>' % tag)
                 else:
                     buf.append('>')
-                yield Markup(u''.join(buf))
+                yield _emit(kind, data, Markup(''.join(buf)))
 
             elif kind is END:
-                yield Markup('</%s>' % data)
+                yield _emit(kind, data, Markup('</%s>' % data))
 
             elif kind is TEXT:
                 if in_cdata:
-                    yield data
+                    yield _emit(kind, data, data)
                 else:
-                    yield escape(data, quotes=False)
+                    yield _emit(kind, data, escape(data, quotes=False))
 
             elif kind is COMMENT:
-                yield Markup('<!--%s-->' % data)
+                yield _emit(kind, data, Markup('<!--%s-->' % data))
 
             elif kind is DOCTYPE and not have_doctype:
                 name, pubid, sysid = data
@@ -352,7 +383,7 @@ class XHTMLSerializer(XMLSerializer):
                 if sysid:
                     buf.append(' "%s"')
                 buf.append('>\n')
-                yield Markup(u''.join(buf)) % filter(None, data)
+                yield Markup(''.join(buf)) % tuple([p for p in data if p])
                 have_doctype = True
 
             elif kind is XML_DECL and not have_decl and not drop_xml_decl:
@@ -364,7 +395,7 @@ class XHTMLSerializer(XMLSerializer):
                     standalone = standalone and 'yes' or 'no'
                     buf.append(' standalone="%s"' % standalone)
                 buf.append('?>\n')
-                yield Markup(u''.join(buf))
+                yield Markup(''.join(buf))
                 have_decl = True
 
             elif kind is START_CDATA:
@@ -376,7 +407,7 @@ class XHTMLSerializer(XMLSerializer):
                 in_cdata = False
 
             elif kind is PI:
-                yield Markup('<?%s %s?>' % data)
+                yield _emit(kind, data, Markup('<?%s %s?>' % data))
 
 
 class HTMLSerializer(XHTMLSerializer):
@@ -384,7 +415,7 @@ class HTMLSerializer(XHTMLSerializer):
     
     >>> from genshi.builder import tag
     >>> elem = tag.div(tag.a(href='foo'), tag.br, tag.hr(noshade=True))
-    >>> print ''.join(HTMLSerializer()(elem.generate()))
+    >>> print(''.join(HTMLSerializer()(elem.generate())))
     <div><a href="foo"></a><br><hr noshade></div>
     """
 
@@ -393,7 +424,7 @@ class HTMLSerializer(XHTMLSerializer):
         QName('style'), QName('http://www.w3.org/1999/xhtml}style')
     ])
 
-    def __init__(self, doctype=None, strip_whitespace=True):
+    def __init__(self, doctype=None, strip_whitespace=True, cache=True):
         """Initialize the HTML serializer.
         
         :param doctype: a ``(name, pubid, sysid)`` tuple that represents the
@@ -401,6 +432,9 @@ class HTMLSerializer(XHTMLSerializer):
                         of the generated output
         :param strip_whitespace: whether extraneous whitespace should be
                                  stripped from the output
+        :param cache: whether to cache the text output per event, which
+                      improves performance for repetitive markup
+        :note: Changed in 0.6: The `cache` parameter was added
         """
         super(HTMLSerializer, self).__init__(doctype, False)
         self.filters = [EmptyTagFilter()]
@@ -409,9 +443,10 @@ class HTMLSerializer(XHTMLSerializer):
                                                  self._NOESCAPE_ELEMS))
         self.filters.append(NamespaceFlattener(prefixes={
             'http://www.w3.org/1999/xhtml': ''
-        }))
+        }, cache=cache))
         if doctype:
             self.filters.append(DocTypeInserter(doctype))
+        self.cache = True
 
     def __call__(self, stream):
         boolean_attrs = self._BOOLEAN_ATTRS
@@ -420,11 +455,29 @@ class HTMLSerializer(XHTMLSerializer):
         have_doctype = False
         noescape = False
 
+        cache = {}
+        cache_get = cache.get
+        if self.cache:
+            def _emit(kind, input, output):
+                cache[kind, input] = output
+                return output
+        else:
+            def _emit(kind, input, output):
+                return output
+
         for filter_ in self.filters:
             stream = filter_(stream)
-        for kind, data, pos in stream:
+        for kind, data, _ in stream:
+            output = cache_get((kind, data))
+            if output is not None:
+                yield output
+                if (kind is START or kind is EMPTY) \
+                        and data[0] in noescape_elems:
+                    noescape = True
+                elif kind is END:
+                    noescape = False
 
-            if kind is START or kind is EMPTY:
+            elif kind is START or kind is EMPTY:
                 tag, attrib = data
                 buf = ['<', tag]
                 for attr, value in attrib:
@@ -432,7 +485,7 @@ class HTMLSerializer(XHTMLSerializer):
                         if value:
                             buf += [' ', attr]
                     elif ':' in attr:
-                        if attr == 'xml:lang' and u'lang' not in attrib:
+                        if attr == 'xml:lang' and 'lang' not in attrib:
                             buf += [' lang="', escape(value), '"']
                     elif attr != 'xmlns':
                         buf += [' ', attr, '="', escape(value), '"']
@@ -440,22 +493,22 @@ class HTMLSerializer(XHTMLSerializer):
                 if kind is EMPTY:
                     if tag not in empty_elems:
                         buf.append('</%s>' % tag)
-                yield Markup(u''.join(buf))
+                yield _emit(kind, data, Markup(''.join(buf)))
                 if tag in noescape_elems:
                     noescape = True
 
             elif kind is END:
-                yield Markup('</%s>' % data)
+                yield _emit(kind, data, Markup('</%s>' % data))
                 noescape = False
 
             elif kind is TEXT:
                 if noescape:
-                    yield data
+                    yield _emit(kind, data, data)
                 else:
-                    yield escape(data, quotes=False)
+                    yield _emit(kind, data, escape(data, quotes=False))
 
             elif kind is COMMENT:
-                yield Markup('<!--%s-->' % data)
+                yield _emit(kind, data, Markup('<!--%s-->' % data))
 
             elif kind is DOCTYPE and not have_doctype:
                 name, pubid, sysid = data
@@ -467,11 +520,11 @@ class HTMLSerializer(XHTMLSerializer):
                 if sysid:
                     buf.append(' "%s"')
                 buf.append('>\n')
-                yield Markup(u''.join(buf)) % filter(None, data)
+                yield Markup(''.join(buf)) % tuple([p for p in data if p])
                 have_doctype = True
 
             elif kind is PI:
-                yield Markup('<?%s %s?>' % data)
+                yield _emit(kind, data, Markup('<?%s %s?>' % data))
 
 
 class TextSerializer(object):
@@ -482,23 +535,24 @@ class TextSerializer(object):
     
     >>> from genshi.builder import tag
     >>> elem = tag.div(tag.a('<Hello!>', href='foo'), tag.br)
-    >>> print elem
+    >>> print(elem)
     <div><a href="foo">&lt;Hello!&gt;</a><br/></div>
-    >>> print ''.join(TextSerializer()(elem.generate()))
+    >>> print(''.join(TextSerializer()(elem.generate())))
     <Hello!>
 
     If text events contain literal markup (instances of the `Markup` class),
     that markup is by default passed through unchanged:
     
     >>> elem = tag.div(Markup('<a href="foo">Hello &amp; Bye!</a><br/>'))
-    >>> print elem.generate().render(TextSerializer)
+    >>> print(elem.generate().render(TextSerializer, encoding=None))
     <a href="foo">Hello &amp; Bye!</a><br/>
     
     You can use the ``strip_markup`` to change this behavior, so that tags and
     entities are stripped from the output (or in the case of entities,
     replaced with the equivalent character):
 
-    >>> print elem.generate().render(TextSerializer, strip_markup=True)
+    >>> print(elem.generate().render(TextSerializer, strip_markup=True,
+    ...                              encoding=None))
     Hello & Bye!
     """
 
@@ -556,8 +610,8 @@ class NamespaceFlattener(object):
     ...   <two:item/>
     ... </doc>''')
     >>> for kind, data, pos in NamespaceFlattener()(xml):
-    ...     print kind, repr(data)
-    START (u'doc', Attrs([(u'xmlns', u'NS1'), (u'xmlns:two', u'NS2')]))
+    ...     print('%s %r' % (kind, data))
+    START (u'doc', Attrs([('xmlns', u'NS1'), (u'xmlns:two', u'NS2')]))
     TEXT u'\n  '
     START (u'two:item', Attrs())
     END u'two:item'
@@ -565,22 +619,46 @@ class NamespaceFlattener(object):
     END u'doc'
     """
 
-    def __init__(self, prefixes=None):
+    def __init__(self, prefixes=None, cache=True):
         self.prefixes = {XML_NAMESPACE.uri: 'xml'}
         if prefixes is not None:
             self.prefixes.update(prefixes)
+        self.cache = cache
 
     def __call__(self, stream):
+        cache = {}
+        cache_get = cache.get
+        if self.cache:
+            def _emit(kind, input, output, pos):
+                cache[kind, input] = output
+                return kind, output, pos
+        else:
+            def _emit(kind, input, output, pos):
+                return output
+
         prefixes = dict([(v, [k]) for k, v in self.prefixes.items()])
         namespaces = {XML_NAMESPACE.uri: ['xml']}
         def _push_ns(prefix, uri):
             namespaces.setdefault(uri, []).append(prefix)
             prefixes.setdefault(prefix, []).append(uri)
+            cache.clear()
+        def _pop_ns(prefix):
+            uris = prefixes.get(prefix)
+            uri = uris.pop()
+            if not uris:
+                del prefixes[prefix]
+            if uri not in uris or uri != uris[-1]:
+                uri_prefixes = namespaces[uri]
+                uri_prefixes.pop()
+                if not uri_prefixes:
+                    del namespaces[uri]
+            cache.clear()
+            return uri
 
         ns_attrs = []
         _push_ns_attr = ns_attrs.append
         def _make_ns_attr(prefix, uri):
-            return u'xmlns%s' % (prefix and ':%s' % prefix or ''), uri
+            return 'xmlns%s' % (prefix and ':%s' % prefix or ''), uri
 
         def _gen_prefix():
             val = 0
@@ -590,8 +668,11 @@ class NamespaceFlattener(object):
         _gen_prefix = _gen_prefix().next
 
         for kind, data, pos in stream:
+            output = cache_get((kind, data))
+            if output is not None:
+                yield kind, output, pos
 
-            if kind is START or kind is EMPTY:
+            elif kind is START or kind is EMPTY:
                 tag, attrs = data
 
                 tagname = tag.localname
@@ -600,9 +681,9 @@ class NamespaceFlattener(object):
                     if tagns in namespaces:
                         prefix = namespaces[tagns][-1]
                         if prefix:
-                            tagname = u'%s:%s' % (prefix, tagname)
+                            tagname = '%s:%s' % (prefix, tagname)
                     else:
-                        _push_ns_attr((u'xmlns', tagns))
+                        _push_ns_attr(('xmlns', tagns))
                         _push_ns('', tagns)
 
                 new_attrs = []
@@ -617,10 +698,10 @@ class NamespaceFlattener(object):
                         else:
                             prefix = namespaces[attrns][-1]
                         if prefix:
-                            attrname = u'%s:%s' % (prefix, attrname)
+                            attrname = '%s:%s' % (prefix, attrname)
                     new_attrs.append((attrname, value))
 
-                yield kind, (tagname, Attrs(ns_attrs + new_attrs)), pos
+                yield _emit(kind, data, (tagname, Attrs(ns_attrs + new_attrs)), pos)
                 del ns_attrs[:]
 
             elif kind is END:
@@ -629,8 +710,8 @@ class NamespaceFlattener(object):
                 if tagns:
                     prefix = namespaces[tagns][-1]
                     if prefix:
-                        tagname = u'%s:%s' % (prefix, tagname)
-                yield kind, tagname, pos
+                        tagname = '%s:%s' % (prefix, tagname)
+                yield _emit(kind, data, tagname, pos)
 
             elif kind is START_NS:
                 prefix, uri = data
@@ -641,15 +722,7 @@ class NamespaceFlattener(object):
 
             elif kind is END_NS:
                 if data in prefixes:
-                    uris = prefixes.get(data)
-                    uri = uris.pop()
-                    if not uris:
-                        del prefixes[data]
-                    if uri not in uris or uri != uris[-1]:
-                        uri_prefixes = namespaces[uri]
-                        uri_prefixes.pop()
-                        if not uri_prefixes:
-                            del namespaces[uri]
+                    uri = _pop_ns(data)
                     if ns_attrs:
                         attr = _make_ns_attr(data, uri)
                         if attr in ns_attrs:
